"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable functional/prefer-immutable-types */
const type_utils_1 = require("@typescript-eslint/type-utils");
const utils_1 = require("@typescript-eslint/utils");
const typescript_1 = require("typescript");
const common_1 = require("./common");
/**
 * An ESLint rule to ban partial division.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noPartialDivision = (0, common_1.createRule)({
    name: "no-partial-division",
    meta: {
        type: "problem",
        docs: {
            description: "Bans partial division.",
            recommended: "error",
        },
        messages: {
            errorStringGeneric: "Division is partial. You should wrap it in a wrapper that returns undefined when the denominator is zero.",
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        return {
            // eslint-disable-next-line functional/no-return-void
            BinaryExpression: (node) => {
                // eslint-disable-next-line functional/no-conditional-statements
                if (node.operator !== "/") {
                    // Binary expressions other than division are safe.
                    return;
                }
                // eslint-disable-next-line functional/no-conditional-statements
                if (node.right.type === utils_1.AST_NODE_TYPES.Literal &&
                    node.right.value !== 0 &&
                    node.right.value !== 0n) {
                    // Division by a literal that isn't zero is safe.
                    return;
                }
                // eslint-disable-next-line functional/no-conditional-statements
                if (node.right.type === utils_1.AST_NODE_TYPES.Identifier) {
                    const denominatorNode = parserServices.esTreeNodeToTSNodeMap.get(node.right);
                    const denominatorNodeType = checker.getTypeAtLocation(denominatorNode);
                    // eslint-disable-next-line functional/no-conditional-statements
                    if ((0, type_utils_1.isTypeFlagSet)(denominatorNodeType, typescript_1.TypeFlags.NumberLike) &&
                        denominatorNodeType.isNumberLiteral() &&
                        denominatorNodeType.value !== 0) {
                        // Division by a number type that is known to be non-zero is safe.
                        return;
                    }
                    // TODO find a way to get ahold of a BigIntLiteralType without type assertions.
                    // There is no equivalent of `isNumberLiteral()` for bigints.
                    // `isLiteral()` returns false so isn't useful.
                    const bigIntBase10Value = 
                    // eslint-disable-next-line @typescript-eslint/consistent-type-assertions, @typescript-eslint/no-unnecessary-condition
                    denominatorNodeType?.value
                        ?.base10Value;
                    // eslint-disable-next-line functional/no-conditional-statements
                    if (bigIntBase10Value !== undefined && bigIntBase10Value !== "0") {
                        // Division by a bigint type that is known to be non-zero is safe.
                        return;
                    }
                }
                // All other division is not provably safe.
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                });
            },
        };
    },
    defaultOptions: [],
});
exports.default = noPartialDivision;
